import UIKit
import Flutter
import Vision
import SwiftUI

@available(iOS 15.0, *)
@UIApplicationMain
@objc class AppDelegate: FlutterAppDelegate {
    
    private var segmentationRequest: VNGeneratePersonSegmentationRequest?
    
    override func application(
        _ application: UIApplication,
        didFinishLaunchingWithOptions launchOptions: [UIApplication.LaunchOptionsKey: Any]?
    ) -> Bool {
        let controller = window?.rootViewController as! FlutterViewController

        segmentationRequest = VNGeneratePersonSegmentationRequest()
        segmentationRequest?.qualityLevel = .accurate
        segmentationRequest?.outputPixelFormat = kCVPixelFormatType_OneComponent8

        let methodChannel = FlutterMethodChannel(
            name: "methodChannel.localRembg",
            binaryMessenger: controller.binaryMessenger
        )

        methodChannel.setMethodCallHandler { [self] (call, result) in
            if call.method == "removeBackground" {
                if let imagePath = call.arguments as? String {
                    if let image = UIImage(contentsOfFile: imagePath) {
                        applyFilter(image: image) { resultImage in
                            if let imageData = resultImage?.pngData() {
                                result(["status": 1, "imageBytes": FlutterStandardTypedData(bytes: imageData)])
                            } else {
                                result(["status": 0, "message": "Unable to process image"])
                            }
                        }
                    } else {
                        result(["status": 0, "message": "Unable to load image from file"])
                    }
                } else {
                    result(["status": 0, "message": "Image path not provided"])
                }
            } else {
                result(FlutterMethodNotImplemented)
            }
        }
        
        GeneratedPluginRegistrant.register(with: self)
        return super.application(application, didFinishLaunchingWithOptions: launchOptions)
    }
    
    func applyFilter(image: UIImage ,completion: @escaping (UIImage?) -> Void) {
        guard let originalCG = image.cgImage, let segmentationRequest = self.segmentationRequest else {
            return completion(nil)
        }
        
        let requiredSize: CGFloat = 600.0
        
        if image.size.width < requiredSize || image.size.height < requiredSize {
            completion(nil)
            return
        }
        
        let newWidth: CGFloat = 600.0
        let newHeight: CGFloat = 600.0
        
        let resizedImage = resizeImage(image: image, targetSize: CGSize(width: newWidth, height: newHeight))
        
        let handler = VNImageRequestHandler(cgImage: resizedImage.cgImage!)
        
        do {
            try handler.perform([segmentationRequest])

            guard let maskPixelBuffer = segmentationRequest.results?.first?.pixelBuffer else {
                return completion(nil)
            }

            let maskImage = CGImage.create(pixelBuffer: maskPixelBuffer)

            return applyBackgroundMask(maskImage, image: resizedImage, completion: completion)
        } catch {
            return completion(nil)
        }
    }

    private func resizeImage(image: UIImage, targetSize: CGSize) -> UIImage {
        let size = image.size

        let widthRatio  = targetSize.width / size.width
        let heightRatio = targetSize.height / size.height

        let scaleFactor = min(widthRatio, heightRatio)

        let scaledWidth  = size.width * scaleFactor
        let scaledHeight = size.height * scaleFactor

        UIGraphicsBeginImageContextWithOptions(CGSize(width: scaledWidth, height: scaledHeight), false, 0.0)
        image.draw(in: CGRect(x: 0, y: 0, width: scaledWidth, height: scaledHeight))
        let newImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()

        return newImage ?? UIImage()
    }
}
